{% load i18n %}

<!-- Nav tabs -->
<ul class="nav nav-tabs" role="tablist">
    <li role="presentation">
        <a href="#assets" aria-controls="assets" role="tab" data-toggle="tab">
            {% translate 'Assets' %}
            {% if is_assets_updating %}&nbsp;<i class="fas fa-exclamation-circle" title="{% translate 'Update in progress...' %}"></i>{% endif %}
        </a>
    </li>
    <li role="presentation"><a href="#attributes" aria-controls="attributes" role="tab" data-toggle="tab">{% translate 'Attributes' %}</a></li>
    <li role="presentation"><a href="#bio" aria-controls="bio" role="tab" data-toggle="tab">{% translate 'Bio' %}</a></li>
    <li role="presentation"><a href="#contacts" aria-controls="contacts" role="tab" data-toggle="tab">{% translate 'Contacts' %}</a></li>
    <li role="presentation"><a href="#contracts" aria-controls="contracts" role="tab" data-toggle="tab">{% translate 'Contracts' %}</a></li>
    {% comment %} <li role="presentation"><a href="#history" aria-controls="history" role="tab" data-toggle="tab">{% translate 'Corporations' %}</a></li> {% endcomment %}

    <li class="dropdown active" role="presentation" role="tab" aria-controls="clones">
        <a href="#" id="myTabDrop2" class="dropdown-toggle" data-toggle="dropdown">{% translate 'Clones' %}&nbsp;<b class="caret"></b></a>
        <ul class="dropdown-menu" role="menu" aria-labelledby="myTabDrop2">
        <li class=""><a href="#implants" tabindex="-1" data-toggle="tab">{% translate 'Implants' %}</a></li>
        <li class=""><a href="#clones" tabindex="-1" data-toggle="tab">{% translate 'Jump Clones' %}</a></li>
        </ul>
    </li>

    <li role="presentation"><a href="#fw_stats" aria-controls="fw_stats" role="tab" data-toggle="tab">{% translate 'Faction Warfare' %}</a></li>
    <li role="presentation"><a href="#loyalty" aria-controls="loyalty" role="tab" data-toggle="tab">{% translate 'Loyalty' %}</a></li>
    <li role="presentation"><a href="#mail" aria-controls="mail" role="tab" data-toggle="tab">{% translate 'Mails' %}</a></li>
    <li role="presentation"><a href="#mining_ledger" aria-controls="mail" role="tab" data-toggle="tab">{% translate 'Mining' %}</a></li>

    <li class="dropdown active" role="presentation" role="tab" aria-controls="clones">
        <a href="#" id="myTabDrop2" class="dropdown-toggle" data-toggle="dropdown">{% translate 'Skills' %}&nbsp;<b class="caret"></b></a>
        <ul class="dropdown-menu" role="menu" aria-labelledby="myTabDrop2">
        <li class=""><a href="#skillqueue" tabindex="-1" data-toggle="tab">{% translate 'Skill Queue' %}</a></li>
        <li class=""><a href="#skill_sets" tabindex="-1" data-toggle="tab">{% translate 'Skill Sets' %}</a></li>
        <li class=""><a href="#skills" tabindex="-1" data-toggle="tab">{% translate 'Skills' %}</a></li>
        </ul>
    </li>

    <li class="dropdown active" role="presentation" role="tab" aria-controls="wallet">
        <a href="#" id="myTabDrop1" class="dropdown-toggle" data-toggle="dropdown">{% translate 'Wallet' %}&nbsp;<b class="caret"></b></a>
        <ul class="dropdown-menu" role="menu" aria-labelledby="myTabDrop1">
        <li class=""><a href="#wallet_journal" tabindex="-1" data-toggle="tab">{% translate 'Transactions' %}</a></li>
        <li class=""><a href="#wallet_transactions" tabindex="-1" data-toggle="tab">{% translate 'Market Transactions' %}</a></li>
        </ul>
    </li>
</ul>
