# Generated by Django 4.0.10 on 2023-04-16 15:57

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("memberaudit", "0005_add_fw_stats"),
    ]

    operations = [
        migrations.AlterField(
            model_name="character",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, db_index=True, verbose_name="created at"
            ),
        ),
        migrations.AlterField(
            model_name="character",
            name="eve_character",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="memberaudit_character",
                to="eveonline.evecharacter",
                verbose_name="eve character",
            ),
        ),
        migrations.AlterField(
            model_name="character",
            name="is_disabled",
            field=models.BooleanField(
                default=False,
                help_text="Disabled characters are no longer updated from ESI.",
                verbose_name="is disabled",
            ),
        ),
        migrations.AlterField(
            model_name="character",
            name="is_shared",
            field=models.BooleanField(
                default=False,
                help_text="Shared characters can be viewed by recruiters",
                verbose_name="is shared",
            ),
        ),
        migrations.AlterField(
            model_name="character",
            name="mailing_lists",
            field=models.ManyToManyField(
                related_name="characters",
                to="memberaudit.mailentity",
                verbose_name="mailing lists",
            ),
        ),
        migrations.AlterField(
            model_name="skillset",
            name="description",
            field=models.TextField(blank=True, verbose_name="description"),
        ),
        migrations.AlterField(
            model_name="skillset",
            name="is_visible",
            field=models.BooleanField(
                db_index=True,
                default=True,
                help_text="Non visible skill sets are not shown to users on their character sheet and used for audit purposes only.",
                verbose_name="is visible",
            ),
        ),
        migrations.AlterField(
            model_name="skillset",
            name="name",
            field=models.CharField(max_length=100, unique=True, verbose_name="name"),
        ),
        migrations.AlterField(
            model_name="skillset",
            name="ship_type",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="Ship type is used for visual presentation only. All skill requirements must be explicitly defined.",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="memberaudit.eveshiptype",
                verbose_name="ship type",
            ),
        ),
        migrations.AlterField(
            model_name="skillsetgroup",
            name="description",
            field=models.TextField(blank=True, verbose_name="description"),
        ),
        migrations.AlterField(
            model_name="skillsetgroup",
            name="is_active",
            field=models.BooleanField(
                db_index=True,
                default=True,
                help_text="Whether this skill set group is in active use",
                verbose_name="is active",
            ),
        ),
        migrations.AlterField(
            model_name="skillsetgroup",
            name="is_doctrine",
            field=models.BooleanField(
                db_index=True,
                default=False,
                help_text="This enables a skill set group to show up correctly in doctrine reports",
                verbose_name="is doctrine",
            ),
        ),
        migrations.AlterField(
            model_name="skillsetgroup",
            name="name",
            field=models.CharField(max_length=100, unique=True, verbose_name="name"),
        ),
        migrations.AlterField(
            model_name="skillsetgroup",
            name="skill_sets",
            field=models.ManyToManyField(
                related_name="groups",
                to="memberaudit.skillset",
                verbose_name="skill sets",
            ),
        ),
        migrations.AlterField(
            model_name="skillsetskill",
            name="recommended_level",
            field=models.PositiveIntegerField(
                blank=True,
                default=None,
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(1),
                    django.core.validators.MaxValueValidator(5),
                ],
                verbose_name="recommended level",
            ),
        ),
        migrations.AlterField(
            model_name="skillsetskill",
            name="required_level",
            field=models.PositiveIntegerField(
                blank=True,
                default=None,
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(1),
                    django.core.validators.MaxValueValidator(5),
                ],
                verbose_name="required level",
            ),
        ),
        migrations.AlterField(
            model_name="skillsetskill",
            name="skill_set",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="skills",
                to="memberaudit.skillset",
                verbose_name="skills",
            ),
        ),
    ]
