# Generated by Django 4.0.9 on 2023-03-21 22:30

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveuniverse", "0007_evetype_description"),
        ("memberaudit", "0004_character_is_disabled"),
    ]

    operations = [
        migrations.AlterField(
            model_name="characterupdatestatus",
            name="section",
            field=models.CharField(
                choices=[
                    ("assets", "assets"),
                    ("character_details", "character details"),
                    ("contacts", "contacts"),
                    ("contracts", "contracts"),
                    ("corporation_history", "corporation history"),
                    ("fw_stats", "faction warfare statistics"),
                    ("implants", "implants"),
                    ("jump_clones", "jump clones"),
                    ("location", "location"),
                    ("loyalty", "loyalty"),
                    ("mails", "mails"),
                    ("mining_ledger", "mining ledger"),
                    ("online_status", "online status"),
                    ("ship", "ship"),
                    ("skills", "skills"),
                    ("skill_queue", "skill queue"),
                    ("skill_sets", "skill sets"),
                    ("wallet_balance", "wallet balance"),
                    ("wallet_journal", "wallet journal"),
                    ("wallet_transactions", "wallet transactions"),
                    ("attributes", "attributes"),
                ],
                db_index=True,
                max_length=64,
            ),
        ),
        migrations.CreateModel(
            name="CharacterFwStats",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "current_rank",
                    models.PositiveSmallIntegerField(default=None, null=True),
                ),
                ("enlisted_on", models.DateTimeField(default=None, null=True)),
                (
                    "highest_rank",
                    models.PositiveSmallIntegerField(default=None, null=True),
                ),
                ("kills_last_week", models.PositiveIntegerField(db_index=True)),
                ("kills_total", models.PositiveIntegerField(db_index=True)),
                ("kills_yesterday", models.PositiveIntegerField(db_index=True)),
                (
                    "victory_points_last_week",
                    models.PositiveIntegerField(db_index=True),
                ),
                ("victory_points_total", models.PositiveIntegerField(db_index=True)),
                (
                    "victory_points_yesterday",
                    models.PositiveIntegerField(db_index=True),
                ),
                (
                    "character",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fw_stats",
                        to="memberaudit.character",
                    ),
                ),
                (
                    "faction",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        to="eveuniverse.evefaction",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]
