
@media all {
    /* User Compliance Tab */
    #tab_user_compliance tbody td:nth-child(1) {
        white-space: nowrap;
    }

    #tab_user_compliance tbody td:nth-child(6) {
        text-align: right;
    }
    /* Corporation Compliance Tab */
    #tab_corporation_compliance tbody td:nth-child(1) {
        white-space: nowrap;
    }

    #tab_corporation_compliance tbody td:nth-child(2) {
        text-align: right;
    }

    #tab_corporation_compliance tbody td:nth-child(3) {
        text-align: right;
    }

    #tab_corporation_compliance tbody td:nth-child(4) {
        text-align: right;
    }
}
