{% load i18n %}

<div role="tabpanel" class="tab-pane" id="implants">
    {% if has_implants %}
        <div class="table-responsive">
            <table class="table table-striped table-width-fix" id="tab_implants">
                <thead></thead>
                <tbody></tbody>
            </table>
        </div>
    {% else %}
        <p>{% translate "No implants" %}</p>
    {% endif %}
    {% include 'memberaudit/partials/character_viewer/tabs/last_updated.html' with update_section=last_updates.implants %}
</div>
